import json
import os

import click

from uipath._cli._utils._console import ConsoleLogger
from uipath._utils.constants import UIPATH_CONFIG_FILE


def auto_discover_entrypoint() -> str:
    """Auto-discover entrypoint from config file.

    Returns:
        Path to the entrypoint

    Raises:
        ValueError: If no entrypoint found or multiple entrypoints exist
    """
    console = ConsoleLogger()

    if not os.path.isfile(UIPATH_CONFIG_FILE):
        raise ValueError(
            f"File '{UIPATH_CONFIG_FILE}' not found. Please run 'uipath init'."
        )

    with open(UIPATH_CONFIG_FILE, "r", encoding="utf-8") as f:
        uipath_config = json.loads(f.read())

    entrypoints = uipath_config.get("entryPoints", [])

    if not entrypoints:
        raise ValueError(
            f"No entrypoints found in {UIPATH_CONFIG_FILE}. Please run 'uipath init'."
        )

    if len(entrypoints) > 1:
        entrypoint_paths = [ep.get("filePath") for ep in entrypoints]
        raise ValueError(
            f"Multiple entrypoints found: {entrypoint_paths}. "
            f"Please specify which entrypoint to use."
        )

    entrypoint = entrypoints[0].get("filePath")
    console.info(
        f"Auto-discovered agent entrypoint: {click.style(entrypoint, fg='cyan')}"
    )
    return entrypoint



# type: ignore
import importlib.resources
import json
import os
import shutil
import uuid
from pathlib import Path
from typing import Any, Dict, Optional

import click

from .._utils.constants import ENV_TELEMETRY_ENABLED
from ..telemetry import track
from ..telemetry._constants import _PROJECT_KEY, _TELEMETRY_CONFIG_FILE
from ._utils._console import ConsoleLogger
from ._utils._input_args import generate_args
from ._utils._parse_ast import generate_bindings_json
from .middlewares import Middlewares

console = ConsoleLogger()

CONFIG_PATH = "uipath.json"


def create_telemetry_config_file(target_directory: str) -> None:
    """Create telemetry file if telemetry is enabled.

    Args:
        target_directory: The directory where the .uipath folder should be created.
    """
    telemetry_enabled = os.getenv(ENV_TELEMETRY_ENABLED, "true").lower() == "true"

    if not telemetry_enabled:
        return

    uipath_dir = os.path.join(target_directory, ".uipath")
    telemetry_file = os.path.join(uipath_dir, _TELEMETRY_CONFIG_FILE)

    if os.path.exists(telemetry_file):
        return

    os.makedirs(uipath_dir, exist_ok=True)
    telemetry_data = {
        _PROJECT_KEY: os.getenv("UIPATH_PROJECT_ID", None) or str(uuid.uuid4())
    }

    with open(telemetry_file, "w") as f:
        json.dump(telemetry_data, f, indent=4)


def generate_env_file(target_directory):
    env_path = os.path.join(target_directory, ".env")

    if not os.path.exists(env_path):
        relative_path = os.path.relpath(env_path, target_directory)
        with open(env_path, "w"):
            pass
        console.success(f" Created '{relative_path}' file.")


def generate_agents_md(target_directory: str) -> None:
    """Generate AGENTS.md file from the packaged resource.

    Args:
        target_directory: The directory where AGENTS.md should be created.
    """
    target_path = os.path.join(target_directory, "AGENTS.md")

    # Skip if file already exists
    if os.path.exists(target_path):
        console.info("Skipping 'AGENTS.md' creation as it already exists.")
        return

    try:
        # Get the resource path using importlib.resources
        source_path = importlib.resources.files("uipath._resources").joinpath(
            "AGENTS.md"
        )

        # Copy the file to the target directory
        with importlib.resources.as_file(source_path) as s_path:
            shutil.copy(s_path, target_path)

        console.success(" Created 'AGENTS.md' file.")
    except Exception as e:
        console.warning(f"Could not create AGENTS.md: {e}")


def get_existing_settings(config_path: str) -> Optional[Dict[str, Any]]:
    """Read existing settings from uipath.json if it exists.

    Args:
        config_path: Path to the uipath.json file.

    Returns:
        The settings dictionary if it exists, None otherwise.
    """
    if not os.path.exists(config_path):
        return None

    try:
        with open(config_path, "r") as config_file:
            existing_config = json.load(config_file)
            return existing_config.get("settings")
    except (json.JSONDecodeError, IOError):
        return None


def get_user_script(directory: str, entrypoint: Optional[str] = None) -> Optional[str]:
    """Find the Python script to process."""
    if entrypoint:
        script_path = os.path.join(directory, entrypoint)
        if not os.path.isfile(script_path):
            console.error(
                f"The {entrypoint} file does not exist in the current directory."
            )
            return None
        return script_path

    python_files = [f for f in os.listdir(directory) if f.endswith(".py")]

    if not python_files:
        console.error(
            "No python files found in the current directory.\nPlease specify the entrypoint: `uipath init <entrypoint_path>`"
        )
        return None
    elif len(python_files) == 1:
        return os.path.join(directory, python_files[0])
    else:
        console.error(
            "Multiple python files found in the current directory.\nPlease specify the entrypoint: `uipath init <entrypoint_path>`"
        )
        return None


def write_config_file(config_data: Dict[str, Any]) -> None:
    existing_settings = get_existing_settings(CONFIG_PATH)
    if existing_settings is not None:
        config_data["settings"] = existing_settings

    with open(CONFIG_PATH, "w") as config_file:
        json.dump(config_data, config_file, indent=4)

    return CONFIG_PATH


@click.command()
@click.argument("entrypoint", required=False, default=None)
@click.option(
    "--infer-bindings/--no-infer-bindings",
    is_flag=True,
    required=False,
    default=True,
    help="Infer bindings from the script.",
)
@track
def init(entrypoint: str, infer_bindings: bool) -> None:
    """Create uipath.json with input/output schemas and bindings."""
    with console.spinner("Initializing UiPath project ..."):
        current_directory = os.getcwd()
        generate_env_file(current_directory)
        create_telemetry_config_file(current_directory)
        generate_agents_md(current_directory)

        result = Middlewares.next(
            "init",
            entrypoint,
            options={"infer_bindings": infer_bindings},
            write_config=write_config_file,
        )

        if result.error_message:
            console.error(
                result.error_message, include_traceback=result.should_include_stacktrace
            )

        if result.info_message:
            console.info(result.info_message)

        if not result.should_continue:
            return

        script_path = get_user_script(current_directory, entrypoint=entrypoint)

        if not script_path:
            return

        try:
            args = generate_args(script_path)

            relative_path = Path(script_path).relative_to(current_directory).as_posix()

            config_data = {
                "entryPoints": [
                    {
                        "filePath": relative_path,
                        "uniqueId": str(uuid.uuid4()),
                        # "type": "process", OR BE doesn't offer json schema support for type: Process
                        "type": "agent",
                        "input": args["input"],
                        "output": args["output"],
                    }
                ]
            }

            # Generate bindings JSON based on the script path
            try:
                if infer_bindings:
                    bindings_data = generate_bindings_json(script_path)
                else:
                    bindings_data = {}
                # Add bindings to the config data
                config_data["bindings"] = bindings_data
            except Exception as e:
                console.warning(f"Warning: Could not generate bindings: {str(e)}")

            config_path = write_config_file(config_data)
            console.success(f"Created '{config_path}' file.")
        except Exception as e:
            console.error(f"Error creating configuration file:\n {str(e)}")



"""Tests for AGENTS.md generation in the init command."""

import os
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from uipath._cli import cli, generate_agents_md


class TestGenerateAgentsMd:
    """Test the generate_agents_md helper function."""

    def test_generate_agents_md_creates_file(self) -> None:
        """Test that AGENTS.md is created successfully."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Mock importlib.resources to return our test file
            mock_source = (
                Path(__file__).parent.parent.parent
                / "src"
                / "uipath"
                / "_resources"
                / "AGENTS.md"
            )

            with (
                patch("uipath._cli.cli_init.importlib.resources.files") as mock_files,
                patch(
                    "uipath._cli.cli_init.importlib.resources.as_file"
                ) as mock_as_file,
                patch("uipath._cli.cli_init.console") as mock_console,
            ):
                # Setup mocks
                mock_path = MagicMock()
                mock_files.return_value.joinpath.return_value = mock_path
                mock_as_file.return_value.__enter__.return_value = mock_source
                mock_as_file.return_value.__exit__.return_value = None

                # Run function
                generate_agents_md(temp_dir)

                # Verify console success message
                mock_console.success.assert_called_once_with(
                    " Created 'AGENTS.md' file."
                )

    def test_generate_agents_md_skips_existing_file(self) -> None:
        """Test that existing AGENTS.md is not overwritten."""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create existing file
            agents_path = Path(temp_dir) / "AGENTS.md"
            original_content = "Original content"
            agents_path.write_text(original_content)

            with patch("uipath._cli.cli_init.console") as mock_console:
                # Run function
                generate_agents_md(temp_dir)

                # Verify file wasn't changed
                assert agents_path.read_text() == original_content

                # Verify console info message
                mock_console.info.assert_called_once_with(
                    "Skipping 'AGENTS.md' creation as it already exists."
                )

    def test_generate_agents_md_handles_errors_gracefully(self) -> None:
        """Test that errors are handled gracefully."""
        with tempfile.TemporaryDirectory() as temp_dir:
            with (
                patch("uipath._cli.cli_init.importlib.resources.files") as mock_files,
                patch("uipath._cli.cli_init.console") as mock_console,
            ):
                # Make it raise an exception
                mock_files.side_effect = RuntimeError("Test error")

                # Run function - should not raise
                generate_agents_md(temp_dir)

                # Verify warning was logged
                mock_console.warning.assert_called_once()
                assert "Could not create AGENTS.md: Test error" in str(
                    mock_console.warning.call_args
                )


class TestInitWithAgentsMd:
    """Test the init command with default AGENTS.md creation."""

    def test_init_creates_agents_md_by_default(
        self, runner: CliRunner, temp_dir: str
    ) -> None:
        """Test that AGENTS.md is created by default."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Create a simple Python file
            with open("main.py", "w") as f:
                f.write("def main(input): return input")

            # Mock the AGENTS.md source file
            mock_source = (
                Path(__file__).parent.parent.parent
                / "src"
                / "uipath"
                / "_resources"
                / "AGENTS.md"
            )

            with (
                patch("uipath._cli.cli_init.importlib.resources.files") as mock_files,
                patch(
                    "uipath._cli.cli_init.importlib.resources.as_file"
                ) as mock_as_file,
            ):
                # Setup mocks
                mock_path = MagicMock()
                mock_files.return_value.joinpath.return_value = mock_path

                # Check if the actual AGENTS.md exists, if so use it
                if mock_source.exists():
                    mock_as_file.return_value.__enter__.return_value = mock_source
                else:
                    # Create a temp file to copy
                    temp_agents = Path(temp_dir) / "temp_agents.md"
                    temp_agents.write_text("Test AGENTS.md content")
                    mock_as_file.return_value.__enter__.return_value = temp_agents

                mock_as_file.return_value.__exit__.return_value = None

                # Run init (AGENTS.md should be created by default)
                result = runner.invoke(cli, ["init"])

                assert result.exit_code == 0
                assert " Created 'AGENTS.md' file." in result.output
                assert os.path.exists("AGENTS.md")

    def test_init_does_not_overwrite_existing_agents_md(
        self, runner: CliRunner, temp_dir: str
    ) -> None:
        """Test that existing AGENTS.md is not overwritten."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Create a simple Python file
            with open("main.py", "w") as f:
                f.write("def main(input): return input")

            # Create existing AGENTS.md
            original_content = "Original AGENTS.md content"
            with open("AGENTS.md", "w") as f:
                f.write(original_content)

            # Run init (AGENTS.md creation is now default)
            result = runner.invoke(cli, ["init"])

            assert result.exit_code == 0
            assert "Skipping 'AGENTS.md' creation" in result.output

            # Verify content wasn't changed
            with open("AGENTS.md", "r") as f:
                assert f.read() == original_content



import json

import inflection


def _convert_to_snake_case(old_dict):
    new_dict = {}
    for key, value in old_dict.items():
        new_key = inflection.underscore(key)
        new_dict[new_key] = value
    return new_dict


def _convert_to_camel_case(old_dict):
    new_dict = {}
    for key, value in old_dict.items():
        if isinstance(value, list):
            new_value = [
                _convert_to_camel_case(v.__dict__) if hasattr(v, "__dict__") else v
                for v in value
            ]
        elif hasattr(value, "__dict__"):
            new_value = _convert_to_camel_case(value.__dict__)
        else:
            new_value = value
        new_key = inflection.camelize(key, False)
        new_dict[new_key] = new_value
    return new_dict


class InputOutput:
    def __init__(self, type=None, properties=None, required=None):
        self.type = type
        self.properties = properties if properties else {}
        self.required = required if required else []


class EntryPoint:
    def __init__(
        self, file_path=None, unique_id=None, type=None, input=None, output=None
    ):
        self.file_path = file_path
        self.unique_id = unique_id
        self.type = type
        self.input = (
            InputOutput(**_convert_to_snake_case(input))
            if isinstance(input, dict)
            else input
        )
        self.output = (
            InputOutput(**_convert_to_snake_case(output))
            if isinstance(output, dict)
            else output
        )

    def __str__(self):
        data = self.__dict__
        camel_cased_data = _convert_to_camel_case(data)
        return json.dumps(camel_cased_data, separators=(",", ":"))


class Bindings:
    def __init__(self, version=None, resources=None):
        self.version = version
        self.resources = resources if resources else []


class Settings:
    def __init__(
        self,
        file_extensions_included=None,
        files_included=None,
        files_excluded=None,
        directories_excluded=None,
    ):
        self.file_extensions_included = (
            file_extensions_included if file_extensions_included else []
        )
        self.files_included = files_included if files_included else []
        self.files_excluded = files_excluded if files_excluded else []
        self.directories_excluded = directories_excluded if directories_excluded else []


class UiPathJson:
    def __init__(self, entry_points=None, bindings=None, settings=None):
        self.entry_points = (
            [EntryPoint(**ep) if isinstance(ep, dict) else ep for ep in entry_points]
            if entry_points
            else []
        )
        self.bindings = (
            Bindings(**bindings)
            if isinstance(bindings, dict)
            else bindings
            if bindings
            else []
        )
        self.settings = (
            Settings(**settings)
            if isinstance(settings, dict)
            else settings
            if settings
            else []
        )

    @classmethod
    def from_json(cls, json_str):
        data = json.loads(json_str)
        entry_points = [
            _convert_to_snake_case(ep) for ep in data.get("entryPoints", [])
        ]
        bindings = _convert_to_snake_case(data.get("bindings", {}))
        settings = _convert_to_snake_case(data.get("settings", {}))
        return cls(entry_points, bindings, settings)

    def to_json(self):
        data = self.__dict__
        camel_cased_data = _convert_to_camel_case(data)
        return json.dumps(camel_cased_data, separators=(",", ":"))



"""Mocking interface."""

import logging
from contextvars import ContextVar
from typing import Any, Callable, Optional

from uipath._cli._evals._models._evaluation_set import EvaluationItem
from uipath._cli._evals.mocks.mocker import Mocker, UiPathNoMockFoundError
from uipath._cli._evals.mocks.mocker_factory import MockerFactory

evaluation_context: ContextVar[Optional[EvaluationItem]] = ContextVar(
    "evaluation", default=None
)

mocker_context: ContextVar[Optional[Mocker]] = ContextVar("mocker", default=None)

logger = logging.getLogger(__name__)


def set_evaluation_item(item: EvaluationItem) -> None:
    """Set an evaluation item within an evaluation set."""
    evaluation_context.set(item)
    try:
        if item.mocking_strategy:
            mocker_context.set(MockerFactory.create(item))
        else:
            mocker_context.set(None)
    except Exception:
        logger.warning(f"Failed to create mocker for evaluation {item.name}")
        mocker_context.set(None)


async def get_mocked_response(
    func: Callable[[Any], Any], params: dict[str, Any], *args, **kwargs
) -> Any:
    """Get a mocked response."""
    mocker = mocker_context.get()
    if mocker is None:
        raise UiPathNoMockFoundError()
    else:
        return await mocker.response(func, params, *args, **kwargs)



import asyncio
import json
import logging
import uuid
from collections import defaultdict
from pathlib import Path
from time import time
from typing import Any, Dict, Generic, List, Optional, Sequence, TypeVar

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

from ..._events._event_bus import EventBus
from ..._events._events import (
    EvalItemExceptionDetails,
    EvalRunCreatedEvent,
    EvalRunUpdatedEvent,
    EvalSetRunCreatedEvent,
    EvalSetRunUpdatedEvent,
    EvaluationEvents,
)
from ...eval.evaluators import BaseEvaluator
from ...eval.models import EvaluationResult
from ...eval.models.models import AgentExecution, EvalItemResult
from .._runtime._contracts import (
    UiPathBaseRuntime,
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)
from .._runtime._logging import ExecutionLogHandler
from .._utils._eval_set import EvalHelpers
from ._evaluator_factory import EvaluatorFactory
from ._models._evaluation_set import EvaluationItem, EvaluationSet
from ._models._exceptions import EvaluationRuntimeException
from ._models._output import (
    EvaluationResultDto,
    EvaluationRunResult,
    EvaluationRunResultDto,
    UiPathEvalOutput,
    UiPathEvalRunExecutionOutput,
)
from .mocks.mocks import set_evaluation_item

T = TypeVar("T", bound=UiPathBaseRuntime)
C = TypeVar("C", bound=UiPathRuntimeContext)


class ExecutionSpanExporter(SpanExporter):
    """Custom exporter that stores spans grouped by execution ids."""

    def __init__(self):
        # { execution_id -> list of spans }
        self._spans: Dict[str, List[ReadableSpan]] = defaultdict(list)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            if span.attributes is not None:
                exec_id = span.attributes.get("execution.id")
                if exec_id is not None and isinstance(exec_id, str):
                    self._spans[exec_id].append(span)

        return SpanExportResult.SUCCESS

    def get_spans(self, execution_id: str) -> List[ReadableSpan]:
        """Retrieve spans for a given execution id."""
        return self._spans.get(execution_id, [])

    def clear(self, execution_id: Optional[str] = None) -> None:
        """Clear stored spans for one or all executions."""
        if execution_id:
            self._spans.pop(execution_id, None)
        else:
            self._spans.clear()

    def shutdown(self) -> None:
        self.clear()


class ExecutionLogsExporter:
    """Custom exporter that stores multiple execution log handlers."""

    def __init__(self):
        self._log_handlers: dict[str, ExecutionLogHandler] = {}

    def register(self, execution_id: str, handler: ExecutionLogHandler) -> None:
        self._log_handlers[execution_id] = handler

    def get_logs(self, execution_id: str) -> list[logging.LogRecord]:
        """Clear stored spans for one or all executions."""
        log_handler = self._log_handlers.get(execution_id)
        return log_handler.buffer if log_handler else []

    def clear(self, execution_id: Optional[str] = None) -> None:
        """Clear stored spans for one or all executions."""
        if execution_id:
            self._log_handlers.pop(execution_id, None)
        else:
            self._log_handlers.clear()

    def flush_logs(self, execution_id: str, target_handler: logging.Handler) -> None:
        log_handler = self._log_handlers.get(execution_id)
        if log_handler:
            log_handler.flush_execution_logs(target_handler)


class UiPathEvalContext(UiPathRuntimeContext):
    """Context used for evaluation runs."""

    no_report: Optional[bool] = False
    workers: Optional[int] = 1
    eval_set: Optional[str] = None
    eval_ids: Optional[List[str]] = None


class UiPathEvalRuntime(UiPathBaseRuntime, Generic[T, C]):
    """Specialized runtime for evaluation runs, with access to the factory."""

    def __init__(
        self,
        context: UiPathEvalContext,
        factory: UiPathRuntimeFactory[T, C],
        event_bus: EventBus,
    ):
        super().__init__(context)
        self.context: UiPathEvalContext = context
        self.factory: UiPathRuntimeFactory[T, C] = factory
        self.event_bus: EventBus = event_bus
        self.span_exporter: ExecutionSpanExporter = ExecutionSpanExporter()
        self.factory.add_span_exporter(self.span_exporter)
        self.logs_exporter: ExecutionLogsExporter = ExecutionLogsExporter()
        self.execution_id = str(uuid.uuid4())

    @classmethod
    def from_eval_context(
        cls,
        context: UiPathEvalContext,
        factory: UiPathRuntimeFactory[T, C],
        event_bus: EventBus,
    ) -> "UiPathEvalRuntime[T, C]":
        return cls(context, factory, event_bus)

    async def execute(self) -> Optional[UiPathRuntimeResult]:
        if self.context.eval_set is None:
            raise ValueError("eval_set must be provided for evaluation runs")

        event_bus = self.event_bus

        evaluation_set = EvalHelpers.load_eval_set(
            self.context.eval_set, self.context.eval_ids
        )
        evaluators = self._load_evaluators(evaluation_set)

        await event_bus.publish(
            EvaluationEvents.CREATE_EVAL_SET_RUN,
            EvalSetRunCreatedEvent(
                execution_id=self.execution_id,
                entrypoint=self.context.entrypoint or "",
                eval_set_id=evaluation_set.id,
                no_of_evals=len(evaluation_set.evaluations),
                evaluators=evaluators,
            ),
        )

        # Check if parallel execution should be used
        if (
            self.context.workers
            and self.context.workers > 1
            and len(evaluation_set.evaluations) > 1
        ):
            eval_run_result_list = await self._execute_parallel(
                evaluation_set, evaluators, event_bus, self.context.workers
            )
        else:
            eval_run_result_list = await self._execute_sequential(
                evaluation_set, evaluators, event_bus
            )
        results = UiPathEvalOutput(
            evaluation_set_name=evaluation_set.name,
            evaluation_set_results=eval_run_result_list,
        )

        # Computing evaluator averages
        evaluator_averages: Dict[str, float] = defaultdict(float)
        evaluator_count: Dict[str, int] = defaultdict(int)

        for eval_run_result in results.evaluation_set_results:
            for result_dto in eval_run_result.evaluation_run_results:
                evaluator_averages[result_dto.evaluator_id] += result_dto.result.score
                evaluator_count[result_dto.evaluator_id] += 1

        for eval_id in evaluator_averages:
            evaluator_averages[eval_id] = (
                evaluator_averages[eval_id] / evaluator_count[eval_id]
            )

        await event_bus.publish(
            EvaluationEvents.UPDATE_EVAL_SET_RUN,
            EvalSetRunUpdatedEvent(
                execution_id=self.execution_id,
                evaluator_scores=evaluator_averages,
            ),
            wait_for_completion=False,
        )

        self.context.result = UiPathRuntimeResult(
            output={**results.model_dump(by_alias=True)},
            status=UiPathRuntimeStatus.SUCCESSFUL,
        )
        return self.context.result

    async def _execute_sequential(
        self,
        evaluation_set: EvaluationSet,
        evaluators: List[BaseEvaluator[Any]],
        event_bus: EventBus,
    ) -> List[EvaluationRunResult]:
        all_eval_run_result: list[EvaluationRunResult] = []

        for eval_item in evaluation_set.evaluations:
            all_eval_run_result.append(
                await self._execute_eval(eval_item, evaluators, event_bus)
            )

        return all_eval_run_result

    async def _execute_parallel(
        self,
        evaluation_set: EvaluationSet,
        evaluators: List[BaseEvaluator[Any]],
        event_bus: EventBus,
        workers: int,
    ) -> List[EvaluationRunResult]:
        # Create a queue with max concurrency
        queue: asyncio.Queue[tuple[int, EvaluationItem]] = asyncio.Queue(
            maxsize=workers
        )

        # Dictionary to store results with their original indices
        results_dict: Dict[int, EvaluationRunResult] = {}

        # Producer task to fill the queue
        async def producer() -> None:
            for index, eval_item in enumerate(evaluation_set.evaluations):
                await queue.put((index, eval_item))
            # Signal completion by putting None markers
            for _ in range(workers):
                await queue.put(None)  # type: ignore

        # Worker function to process items from the queue
        async def worker(worker_id: int) -> None:
            while True:
                item = await queue.get()

                # Check for termination signal
                if item is None:
                    queue.task_done()
                    break

                index, eval_item = item

                try:
                    # Execute the evaluation
                    result = await self._execute_eval(eval_item, evaluators, event_bus)

                    # Store result with its index to maintain order
                    results_dict[index] = result
                finally:
                    # Mark the task as done
                    queue.task_done()

        # Start producer
        producer_task = asyncio.create_task(producer())

        # Create worker tasks based on workers
        worker_tasks = [asyncio.create_task(worker(i)) for i in range(workers)]

        # Wait for producer and all workers to complete
        await producer_task
        await asyncio.gather(*worker_tasks)

        # Return results in the original order
        return [results_dict[i] for i in range(len(evaluation_set.evaluations))]

    async def _execute_eval(
        self,
        eval_item: EvaluationItem,
        evaluators: List[BaseEvaluator[Any]],
        event_bus: EventBus,
    ) -> EvaluationRunResult:
        set_evaluation_item(eval_item)

        await event_bus.publish(
            EvaluationEvents.CREATE_EVAL_RUN,
            EvalRunCreatedEvent(
                execution_id=self.execution_id,
                eval_item=eval_item,
            ),
        )

        evaluation_run_results = EvaluationRunResult(
            evaluation_name=eval_item.name, evaluation_run_results=[]
        )

        try:
            agent_execution_output = await self.execute_runtime(eval_item)
            evaluation_item_results: list[EvalItemResult] = []

            for evaluator in evaluators:
                evaluation_result = await self.run_evaluator(
                    evaluator=evaluator,
                    execution_output=agent_execution_output,
                    eval_item=eval_item,
                )

                dto_result = EvaluationResultDto.from_evaluation_result(
                    evaluation_result
                )

                evaluation_run_results.evaluation_run_results.append(
                    EvaluationRunResultDto(
                        evaluator_name=evaluator.name,
                        result=dto_result,
                        evaluator_id=evaluator.id,
                    )
                )
                evaluation_item_results.append(
                    EvalItemResult(
                        evaluator_id=evaluator.id,
                        result=evaluation_result,
                    )
                )

            await event_bus.publish(
                EvaluationEvents.UPDATE_EVAL_RUN,
                EvalRunUpdatedEvent(
                    execution_id=self.execution_id,
                    eval_item=eval_item,
                    eval_results=evaluation_item_results,
                    success=not agent_execution_output.result.error,
                    agent_output=agent_execution_output.result.output,
                    agent_execution_time=agent_execution_output.execution_time,
                    spans=agent_execution_output.spans,
                    logs=agent_execution_output.logs,
                ),
                wait_for_completion=False,
            )

        except Exception as e:
            exception_details = EvalItemExceptionDetails(exception=e)

            for evaluator in evaluators:
                evaluation_run_results.evaluation_run_results.append(
                    EvaluationRunResultDto(
                        evaluator_name=evaluator.name,
                        evaluator_id=evaluator.id,
                        result=EvaluationResultDto(score=0),
                    )
                )

            eval_run_updated_event = EvalRunUpdatedEvent(
                execution_id=self.execution_id,
                eval_item=eval_item,
                eval_results=[],
                success=False,
                agent_output={},
                agent_execution_time=0.0,
                exception_details=exception_details,
                spans=[],
                logs=[],
            )
            if isinstance(e, EvaluationRuntimeException):
                eval_run_updated_event.spans = e.spans
                eval_run_updated_event.logs = e.logs
                eval_run_updated_event.exception_details.exception = (  # type: ignore
                    e.root_exception
                )
                eval_run_updated_event.exception_details.runtime_exception = True  # type: ignore

            await event_bus.publish(
                EvaluationEvents.UPDATE_EVAL_RUN,
                eval_run_updated_event,
                wait_for_completion=False,
            )

        return evaluation_run_results

    def _get_and_clear_execution_data(
        self, execution_id: str
    ) -> tuple[List[ReadableSpan], list[logging.LogRecord]]:
        spans = self.span_exporter.get_spans(execution_id)
        self.span_exporter.clear(execution_id)

        logs = self.logs_exporter.get_logs(execution_id)
        self.logs_exporter.clear(execution_id)

        return spans, logs

    async def execute_runtime(
        self, eval_item: EvaluationItem
    ) -> UiPathEvalRunExecutionOutput:
        eval_item_id = eval_item.id
        runtime_context: C = self.factory.new_context(
            execution_id=eval_item_id,
            input_json=eval_item.inputs,
            is_eval_run=True,
            log_handler=self._setup_execution_logging(eval_item_id),
        )
        if runtime_context.execution_id is None:
            raise ValueError("execution_id must be set for eval runs")

        attributes = {
            "evalId": eval_item.id,
            "span_type": "eval",
        }
        if runtime_context.execution_id:
            attributes["execution.id"] = runtime_context.execution_id

        start_time = time()
        try:
            result = await self.factory.execute_in_root_span(
                runtime_context, root_span=eval_item.name, attributes=attributes
            )
        except Exception as e:
            spans, logs = self._get_and_clear_execution_data(
                runtime_context.execution_id
            )
            raise EvaluationRuntimeException(
                spans=spans,
                logs=logs,
                root_exception=e,
            ) from e

        end_time = time()
        spans, logs = self._get_and_clear_execution_data(runtime_context.execution_id)

        if result is None:
            raise ValueError("Execution result cannot be None for eval runs")

        return UiPathEvalRunExecutionOutput(
            execution_time=end_time - start_time,
            spans=spans,
            logs=logs,
            result=result,
        )

    def _setup_execution_logging(self, eval_item_id: str) -> ExecutionLogHandler:
        execution_log_handler = ExecutionLogHandler(eval_item_id)
        self.logs_exporter.register(eval_item_id, execution_log_handler)
        return execution_log_handler

    async def run_evaluator(
        self,
        evaluator: BaseEvaluator[Any],
        execution_output: UiPathEvalRunExecutionOutput,
        eval_item: EvaluationItem,
    ) -> EvaluationResult:
        agent_execution = AgentExecution(
            agent_input=eval_item.inputs,
            agent_output=execution_output.result.output or {},
            agent_trace=execution_output.spans,
            expected_agent_behavior=eval_item.expected_agent_behavior,
        )

        result = await evaluator.evaluate(
            agent_execution=agent_execution,
            # at the moment evaluation_criteria is always the expected output
            evaluation_criteria=eval_item.expected_output,
        )

        return result

    def _load_evaluators(
        self, evaluation_set: EvaluationSet
    ) -> List[BaseEvaluator[Any]]:
        """Load evaluators referenced by the evaluation set."""
        evaluators = []
        evaluators_dir = Path(self.context.eval_set).parent.parent / "evaluators"  # type: ignore
        evaluator_refs = set(evaluation_set.evaluator_refs)
        found_evaluator_ids = set()

        for file in evaluators_dir.glob("*.json"):
            try:
                with open(file, "r", encoding="utf-8") as f:
                    data = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"Invalid JSON in evaluator file '{file}': {str(e)}. "
                    f"Please check the file for syntax errors."
                ) from e

            try:
                evaluator_id = data.get("id")
                if evaluator_id in evaluator_refs:
                    evaluator = EvaluatorFactory.create_evaluator(data)
                    evaluators.append(evaluator)
                    found_evaluator_ids.add(evaluator_id)
            except Exception as e:
                raise ValueError(
                    f"Failed to create evaluator from file '{file}': {str(e)}. "
                    f"Please verify the evaluator configuration."
                ) from e

        missing_evaluators = evaluator_refs - found_evaluator_ids
        if missing_evaluators:
            raise ValueError(
                f"Could not find the following evaluators: {missing_evaluators}"
            )

        return evaluators

    async def cleanup(self) -> None:
        """Cleanup runtime resources."""
        pass

    async def validate(self) -> None:
        """Cleanup runtime resources."""
        pass



import json
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Tuple


class OverwritesManager:
    """Manages overwrites for different resource types and methods.

    This class handles reading and accessing bindings overwrites from a JSON file.
    The overwrites are stored under the 'resourceOverwrites' key, where each key is a
    resource key (e.g., 'asset.MyAssetKeyFromBindingsJson') and the value contains
    'name' and 'folderPath' fields.

    This is a singleton class to ensure only one instance exists throughout the application.
    """

    _instance = None
    _overwrites_file_path: Path = Path("__uipath/uipath.json")
    _runtime_overwrites: Dict[str, Any] = {}

    def __new__(
        cls, overwrites_file_path: Optional[Path] = None
    ) -> "OverwritesManager":
        """Create or return the singleton instance.

        Args:
            overwrites_file_path: Optional path to the overwrites JSON file.
                If not provided, defaults to 'uipath.json' in the project root.

        Returns:
            The singleton instance of OverwritesManager.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._overwrites_file_path = (
                overwrites_file_path or cls._overwrites_file_path
            )
            cls._instance._read_overwrites_file()
        elif (
            overwrites_file_path
            and overwrites_file_path != cls._instance._overwrites_file_path
        ):
            # If a new file path is provided and it's different from the current one,
            # update the path and re-read the file
            cls._instance._overwrites_file_path = overwrites_file_path
            cls._instance._read_overwrites_file()
        return cls._instance

    def _read_overwrites_file(self) -> None:
        """Read the overwrites JSON file and cache the data.

        Raises:
            FileNotFoundError: If the overwrites file doesn't exist.
            json.JSONDecodeError: If the file contains invalid JSON.
        """
        try:
            with open(self._overwrites_file_path, "r") as f:
                data = json.load(f)
                self._runtime_overwrites = (
                    data.get("runtime", {})
                    .get("internalArguments", {})
                    .get("resourceOverwrites", {})
                )
        except FileNotFoundError:
            self._runtime_overwrites = {}

    def get_overwrite(
        self, resource_type: str, resource_name: str, folder_path: Optional[str] = None
    ) -> Optional[Tuple[str, str]]:
        """Get an overwrite value for a specific resource.

        Args:
            resource_type: The type of resource (e.g., 'process', 'asset').
            resource_name: The name of the resource.

        Returns:
            A tuple of (name, folder_path) if found, None otherwise.
        """
        if folder_path:
            key = f"{resource_type}.{resource_name}.{folder_path}"
        else:
            key = f"{resource_type}.{resource_name}"

        if key not in self._runtime_overwrites:
            return None

        overwrite = self._runtime_overwrites[key]
        return (
            overwrite.get("name", resource_name),
            overwrite.get("folderPath", ""),
        )

    def get_and_apply_overwrite(
        self, resource_type: str, resource_name: str, folder_path: Optional[str] = None
    ) -> Tuple[Any, Any]:
        """Get and apply overwrites for a resource, falling back to provided values if no overwrites exist.

        Args:
            resource_type: The type of resource (e.g., 'process', 'asset').
            resource_name: The name of the resource.
            folder_path: Optional folder path to use if no overwrite exists.

        Returns:
            A tuple of (name, folder_path) with overwritten values if available,
            otherwise the original values.
        """
        overwrite = self.get_overwrite(resource_type, resource_name, folder_path)
        if overwrite:
            resource_name, folder_path = overwrite
        return resource_name, folder_path or None


@contextmanager
def read_resource_overwrites(
    resource_type: str,
    resource_name: str,
    folder_path: Optional[str] = None,
    overwrites_file_path: Optional[Path] = None,
) -> Generator[Tuple[str, str], None, None]:
    """Context manager for reading and applying resource overwrites.

    Args:
        resource_type: The type of resource (e.g., 'process', 'asset').
        resource_name: The name of the resource.
        folder_path: Optional folder path to use if no overwrite exists.
        overwrites_file_path: Optional path to the overwrites JSON file.

    Yields:
        A tuple of (name, folder_path) with overwritten values if available,
        otherwise the original values.

    Example:
        ```python
        with read_resource_overwrites("asset", "MyAsset") as (name, folder_path):
            # Use name and folder_path here
            pass
        ```
    """
    manager = OverwritesManager(overwrites_file_path)
    try:
        yield manager.get_and_apply_overwrite(resource_type, resource_name, folder_path)
    finally:
        pass



"""Tool call events."""

from typing import Any, Dict, Optional

from pydantic import BaseModel, ConfigDict, Field

from .content import InlineOrExternal


class UiPathConversationToolCallResult(BaseModel):
    """Represents the result of a tool call execution."""

    timestamp: Optional[str] = None
    value: Optional[InlineOrExternal] = None
    is_error: Optional[bool] = Field(None, alias="isError")
    cancelled: Optional[bool] = None

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationToolCall(BaseModel):
    """Represents a call to an external tool or function within a message."""

    tool_call_id: str = Field(..., alias="toolCallId")
    name: str
    arguments: Optional[InlineOrExternal] = None
    timestamp: Optional[str] = None
    result: Optional[UiPathConversationToolCallResult] = None

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationToolCallStartEvent(BaseModel):
    """Signals the start of a tool call."""

    tool_name: str = Field(..., alias="toolName")
    timestamp: Optional[str] = None
    arguments: Optional[InlineOrExternal] = None
    meta_data: Optional[Dict[str, Any]] = Field(None, alias="metaData")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationToolCallEndEvent(BaseModel):
    """Signals the end of a tool call."""

    timestamp: Optional[str] = None
    result: Optional[Any] = None
    is_error: Optional[bool] = Field(None, alias="isError")
    cancelled: Optional[bool] = None
    meta_data: Optional[Dict[str, Any]] = Field(None, alias="metaData")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationToolCallEvent(BaseModel):
    """Encapsulates the data related to a tool call event."""

    tool_call_id: str = Field(..., alias="toolCallId")
    start: Optional[UiPathConversationToolCallStartEvent] = None
    end: Optional[UiPathConversationToolCallEndEvent] = None
    meta_event: Optional[Dict[str, Any]] = Field(None, alias="metaEvent")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)



# type: ignore
"""CLI command for pulling remote project files from UiPath StudioWeb solution.

This module provides functionality to pull remote project files from a UiPath StudioWeb solution.
It handles:
- File downloads from source_code and evals folders
- Maintaining folder structure locally
- File comparison using hashes
- Interactive confirmation for overwriting files
"""

# type: ignore
import asyncio
import hashlib
import json
import os
from typing import Dict, Set

import click

from ..telemetry import track
from ._utils._console import ConsoleLogger
from ._utils._constants import UIPATH_PROJECT_ID
from ._utils._studio_project import (
    ProjectFile,
    ProjectFolder,
    StudioClient,
    get_folder_by_name,
)

console = ConsoleLogger()


def compute_normalized_hash(content: str) -> str:
    """Compute hash of normalized content.

    Args:
        content: Content to hash

    Returns:
        str: SHA256 hash of the normalized content
    """
    try:
        # Try to parse as JSON to handle formatting
        json_content = json.loads(content)
        normalized = json.dumps(json_content, indent=2)
    except json.JSONDecodeError:
        # Not JSON, normalize line endings
        normalized = content.replace("\r\n", "\n").replace("\r", "\n")

    return hashlib.sha256(normalized.encode("utf-8")).hexdigest()


def collect_files_from_folder(
    folder: ProjectFolder, base_path: str, files_dict: Dict[str, ProjectFile]
) -> None:
    """Recursively collect all files from a folder and its subfolders.

    Args:
        folder: The folder to collect files from
        base_path: Base path for file paths
        files_dict: Dictionary to store collected files
    """
    # Add files from current folder
    for file in folder.files:
        file_path = os.path.join(base_path, file.name)
        files_dict[file_path] = file

    # Recursively process subfolders
    for subfolder in folder.folders:
        subfolder_path = os.path.join(base_path, subfolder.name)
        collect_files_from_folder(subfolder, subfolder_path, files_dict)


async def download_folder_files(
    studio_client: StudioClient,
    folder: ProjectFolder,
    base_path: str,
    processed_files: Set[str],
) -> None:
    """Download files from a folder recursively.

    Args:
        studio_client: Studio client
        folder: The folder to download files from
        base_path: Base path for local file storage
        processed_files: Set to track processed files
    """
    files_dict: Dict[str, ProjectFile] = {}
    collect_files_from_folder(folder, "", files_dict)

    for file_path, remote_file in files_dict.items():
        local_path = os.path.join(base_path, file_path)
        local_dir = os.path.dirname(local_path)

        # Create directory if it doesn't exist
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)

        # Download remote file
        response = await studio_client.download_file_async(remote_file.id)
        remote_content = response.read().decode("utf-8")
        remote_hash = compute_normalized_hash(remote_content)

        if os.path.exists(local_path):
            # Read and hash local file
            with open(local_path, "r", encoding="utf-8") as f:
                local_content = f.read()
                local_hash = compute_normalized_hash(local_content)

            # Compare hashes
            if local_hash != remote_hash:
                styled_path = click.style(str(file_path), fg="cyan")
                console.warning(f"File {styled_path}" + " differs from remote version.")
                response = click.prompt("Do you want to overwrite it? (y/n)", type=str)
                if response.lower() == "y":
                    with open(local_path, "w", encoding="utf-8", newline="\n") as f:
                        f.write(remote_content)
                    console.success(f"Updated {click.style(str(file_path), fg='cyan')}")
                else:
                    console.info(f"Skipped {click.style(str(file_path), fg='cyan')}")
            else:
                console.info(
                    f"File {click.style(str(file_path), fg='cyan')} is up to date"
                )
        else:
            # File doesn't exist locally, create it
            with open(local_path, "w", encoding="utf-8", newline="\n") as f:
                f.write(remote_content)
            console.success(f"Downloaded {click.style(str(file_path), fg='cyan')}")

        processed_files.add(file_path)


@click.command()
@click.argument(
    "root", type=click.Path(exists=True, file_okay=False, dir_okay=True), default="."
)
@track
def pull(root: str) -> None:
    """Pull remote project files from Studio Web Project.

    This command pulls the remote project files from a UiPath Studio Web project.
    It downloads files from the source_code and evals folders, maintaining the
    folder structure locally. Files are compared using hashes before overwriting,
    and user confirmation is required for differing files.

    Args:
        root: The root directory to pull files into

    Environment Variables:
        UIPATH_PROJECT_ID: Required. The ID of the UiPath Studio Web project

    Example:
        $ uipath pull
        $ uipath pull /path/to/project
    """
    if not (project_id := os.getenv(UIPATH_PROJECT_ID, False)):
        console.error("UIPATH_PROJECT_ID environment variable not found.")

    studio_client = StudioClient(project_id)

    with console.spinner("Pulling UiPath project files..."):
        try:
            structure = asyncio.run(studio_client.get_project_structure_async())

            processed_files: Set[str] = set()

            # Process source_code folder
            source_code_folder = get_folder_by_name(structure, "source_code")
            if source_code_folder:
                asyncio.run(
                    download_folder_files(
                        studio_client,
                        source_code_folder,
                        root,
                        processed_files,
                    )
                )
            else:
                console.warning("No source_code folder found in remote project")

            # Process evals folder
            evals_folder = get_folder_by_name(structure, "evals")
            if evals_folder:
                evals_path = os.path.join(root, "evals")
                asyncio.run(
                    download_folder_files(
                        studio_client,
                        evals_folder,
                        evals_path,
                        processed_files,
                    )
                )
            else:
                console.warning("No evals folder found in remote project")

        except Exception as e:
            console.error(f"Failed to pull UiPath project: {str(e)}")



import logging
from dataclasses import dataclass
from typing import Optional

from uipath import UiPath

logger = logging.getLogger(__name__)

sdk = None


def test_assets(sdk: UiPath):
    sdk.assets.retrieve(name="MyAsset")


async def test_llm(sdk: UiPath):
    messages = [
        {"role": "system", "content": "You are a helpful programming assistant."},
        {"role": "user", "content": "How do I read a file in Python?"},
        {"role": "assistant", "content": "You can use the built-in open() function."},
        {"role": "user", "content": "Can you show an example?"},
    ]

    result_openai = await sdk.llm_openai.chat_completions(messages)
    logger.info("LLM OpenAI Response: %s", result_openai.choices[0].message.content)

    result_normalized = await sdk.llm.chat_completions(messages)
    logger.info(
        "LLM Normalized Response: %s", result_normalized.choices[0].message.content
    )


@dataclass
class EchoIn:
    message: str


@dataclass
class EchoOut:
    message: str


async def main(input: EchoIn) -> EchoOut:
    sdk = UiPath()

    await test_llm(sdk)
    return EchoOut(message=input.message)



from typing import Any, Dict

from pydantic import TypeAdapter

from uipath._cli._evals._models._evaluator import (
    EqualsEvaluatorParams,
    Evaluator,
    JsonSimilarityEvaluatorParams,
    LLMEvaluatorParams,
    TrajectoryEvaluatorParams,
)
from uipath._cli._evals._models._evaluator_base_params import EvaluatorBaseParams
from uipath.eval.evaluators import (
    BaseEvaluator,
    ExactMatchEvaluator,
    JsonSimilarityEvaluator,
    LlmAsAJudgeEvaluator,
    TrajectoryEvaluator,
)


class EvaluatorFactory:
    """Factory class for creating evaluator instances based on configuration."""

    @classmethod
    def create_evaluator(cls, data: Dict[str, Any]) -> BaseEvaluator[Any]:
        """Create an evaluator instance from configuration data.

        Args:
            data: Dictionary containing evaluator configuration from JSON file

        Returns:
            Appropriate evaluator instance based on category

        Raises:
            ValueError: If category is unknown or required fields are missing
        """
        # Extract common fields
        name = data.get("name", "")
        if not name:
            raise ValueError("Evaluator configuration must include 'name' field")
        id = data.get("id", "")
        if not id:
            raise ValueError("Evaluator configuration must include 'id' field")

        params: EvaluatorBaseParams = TypeAdapter(Evaluator).validate_python(data)

        match params:
            case EqualsEvaluatorParams():
                return EvaluatorFactory._create_exact_match_evaluator(params)
            case JsonSimilarityEvaluatorParams():
                return EvaluatorFactory._create_json_similarity_evaluator(params)
            case LLMEvaluatorParams():
                return EvaluatorFactory._create_llm_as_judge_evaluator(params)
            case TrajectoryEvaluatorParams():
                return EvaluatorFactory._create_trajectory_evaluator(params)
            case _:
                raise ValueError(f"Unknown evaluator category: {params}")

    @staticmethod
    def _create_exact_match_evaluator(
        params: EqualsEvaluatorParams,
    ) -> ExactMatchEvaluator:
        """Create a deterministic evaluator."""
        return ExactMatchEvaluator(**params.model_dump())

    @staticmethod
    def _create_json_similarity_evaluator(
        params: JsonSimilarityEvaluatorParams,
    ) -> JsonSimilarityEvaluator:
        """Create a deterministic evaluator."""
        return JsonSimilarityEvaluator(**params.model_dump())

    @staticmethod
    def _create_llm_as_judge_evaluator(
        params: LLMEvaluatorParams,
    ) -> LlmAsAJudgeEvaluator:
        """Create an LLM-as-a-judge evaluator."""
        if not params.prompt:
            raise ValueError("LLM evaluator must include 'prompt' field")

        if not params.model:
            raise ValueError("LLM evaluator must include 'model' field")
        if params.model == "same-as-agent":
            raise ValueError(
                "'same-as-agent' model option is not supported by coded agents evaluations. Please select a specific model for the evaluator."
            )

        return LlmAsAJudgeEvaluator(**params.model_dump())

    @staticmethod
    def _create_trajectory_evaluator(
        params: TrajectoryEvaluatorParams,
    ) -> TrajectoryEvaluator:
        """Create a trajectory evaluator."""
        if not params.prompt:
            raise ValueError("Trajectory evaluator must include 'prompt' field")

        if not params.model:
            raise ValueError("LLM evaluator must include 'model' field")
        if params.model == "same-as-agent":
            raise ValueError(
                "'same-as-agent' model option is not supported by coded agents evaluations. Please select a specific model for the evaluator."
            )

        return TrajectoryEvaluator(**params.model_dump())



from uipath import UiPath

uipath = UiPath()


async def async_function() -> None:
    await uipath.assets.retrieve_async("asset_from_retrieve_async")
    await uipath.buckets.retrieve_async("bucket_from_retrieve_async")
    await uipath.context_grounding.retrieve_async("index_from_retrieve_async")
    await uipath.processes.invoke_async("process_name_from_invoke_async")


def include_folder_path() -> None:
    uipath.assets.retrieve("asset_with_folder_path", folder_path="asset_folder_path")
    uipath.buckets.retrieve("bucket_with_folder_path", folder_path="bucket_folder_path")
    uipath.context_grounding.retrieve(
        "index_with_folder_path", folder_path="index_folder_path"
    )
    uipath.processes.invoke(
        "process_with_folder_path", folder_path="process_folder_path"
    )


def main() -> None:
    uipath.assets.retrieve("asset_from_retrieve")
    uipath.buckets.retrieve("bucket_from_retrieve")
    uipath.context_grounding.retrieve("index_from_retrieve")
    uipath.processes.invoke("process_name_from_invoke")
